@@ -1,128 +1,107 @@
 import logging
 import numpy as np
 from numpy import newaxis
-from ..general_functions import strm
 from .. import nnls as this_nnls
+from ..general_functions import strm
+import _nnls
 
 logger = logging.getLogger("pyspecdata.matrix_math")
 
 # {{{ local functions
-def chi(x_vec, val, data_fornnls, K):
-    # appears to not be used?
-    return 0.5 * np.dot(
-        x_vec.T, np.dot(dd_chi(G(x_vec, K), val**2), x_vec)
-    ) - np.dot(x_vec.T, data_fornnls[:, newaxis])
 
-def d_chi(x_vec, val, data_fornnls, K):
-    return np.dot(dd_chi(G(x_vec, K), val**2), x_vec) - data_fornnls[:, newaxis]
 
-def dd_chi(G, val):
-    return G + (val**2) * np.eye(np.shape(G)[0])
+def venk_BRD(initial_α, K_0, mvec, tol=1e-3, maxiter=300):
+    """Wrapper calling the compiled Butler-Reeds-Dawson algorithm."""
+    f, alpha_new = _nnls.venk_brd(initial_α, K_0, mvec, tol, maxiter)
+    return f, alpha_new
 
-def G(x_vec, K):
-    return np.dot(K, np.dot(square_heaviside(x_vec, K), K.T))
 
-def H(product):
-    if product <= 0:
-        return 0
-    if product > 0:
-        return 1
+def venk_nnls(K_0, mvec, l):
+    r"""Use the BRD NNLS routine for a given ``lambda`` value.
 
-def square_heaviside(x_vec, K):
-    K_dim = np.shape(K.T)[0]
-    diag_heavi = np.empty(K_dim)
-    for q in range(K_dim):
-        pull_val = np.dot(K.T[q, :], x_vec)
-        temp = pull_val[0]
-        diag_heavi[q] = H(temp)
-    return np.diag(diag_heavi)
+    Parameters
+    ----------
+    K_0 : ndarray
+        Kernel matrix ``A``.
+    mvec : ndarray
+        Data vector ``b``.
+    l : float
+        Regularization parameter :math:`\lambda`.
 
-def optimize_alpha(input_vec, val, factor, K, tol=1e-6):
-    # if not twoD, then factor should be None
-    alpha_converged = False
-    if factor is None:
-        factor = np.sqrt(input_vec.shape[0])
-    T = np.linalg.inv(dd_chi(G(input_vec, K), val**2))
-    dot_product = np.dot(input_vec.T, np.dot(T, input_vec))
-    ans = dot_product * factor
-    ans = ans / np.linalg.norm(input_vec) / dot_product
-    if abs(ans - val**2) <= tol:
-        logger.debug(strm("ALPHA HAS CONVERGED."))
-        alpha_converged = True
-        return ans, alpha_converged
-    return ans, alpha_converged
-
-def newton_min(input_vec, val, data_fornnls, K):
-    fder = dd_chi(G(input_vec, K), val)
-    fval = d_chi(input_vec, val, data_fornnls, K)
-    return input_vec + np.dot(np.linalg.inv(fder), fval)
-
-def mod_BRD(guess, K, factor, data_fornnls, maxiter=20):
-    smoothing_param = guess
-    alpha_converged = False
-    for iter in range(maxiter):
-        logger.debug(strm("ITERATION NO.", iter))
-        logger.debug(strm("CURRENT LAMBDA", smoothing_param))
-        retval, residual = this_nnls.nnls_regularized(
-            K, data_fornnls, l=smoothing_param
-        )
-        f_vec = retval[:, newaxis]
-        alpha = smoothing_param**2
-        c_vec = np.dot(K, f_vec) - data_fornnls[:, newaxis]
-        c_vec /= -1 * alpha
-        c_update = newton_min(c_vec, smoothing_param, data_fornnls, K)
-        alpha_update, alpha_converged = optimize_alpha(
-            c_update, smoothing_param, factor, K
-        )
-        lambda_update = np.sqrt(alpha_update[0, 0])
-        if alpha_converged:
-            logger.debug(strm("*** OPTIMIZED LAMBDA", lambda_update, "***"))
+    Returns
+    -------
+    tuple(ndarray, ndarray)
+        The solution vector and residual ``A·x - b``.
+    """
+    c = np.ones(K_0.shape[0])
+    for j in range(20):
+        logging.debug(f"venk_nnls iteration {j} for l={l}")
+        # re-run to make c is converged
+        old_c = c.copy()
+        _nnls.venk_nnls(K_0, mvec, c, l**2)
+        logging.debug(f"venk_nnls gave c={c}")
+        if np.linalg.norm(c - old_c) / np.linalg.norm(c) < 1e-5:
             break
-        if not alpha_converged:
-            logger.debug(strm("UPDATED LAMBDA", lambda_update))
-            smoothing_param = lambda_update
-        if iter == maxiter - 1:
-            logger.debug(strm("DID NOT CONVERGE."))
-    return lambda_update
+    f = K_0.T.dot(c)
+    f[f < 0] = 0
+    residual = K_0.dot(f) - mvec
+    return f, residual
+
 
 def demand_real(x, addtxt=""):
     if not x.dtype == np.float64:
         if x.dtype == np.complex128:
             raise ValueError(
-                "you are not allows to pass nnls complex data:\nif it makes sense for you, try yourdata.real.nnls( np.where you now have yourdata.nnls("
+                "you are not allows to pass nnls complex data:\nif it makes"
+                " sense for you, try yourdata.real.nnls( np.where you now have"
+                " yourdata.nnls("
                 + "\n"
                 + addtxt
             )
         else:
             raise ValueError(
-                "I expect double-precision floating point (float64), but you passed me data of dtype "
+                "I expect double-precision floating point (float64), but you"
+                " passed me data of dtype "
                 + str(x.dtype)
                 + "\n"
                 + addtxt
             )
+
+
 # }}}
-def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
-         store_uncompressed_kernel=False):
+def nnls(
+    self,
+    nddata_cls,
+    dimname_list,
+    newaxis_dict,
+    kernel_func,
+    l=0,
+    default_cut=1e-3,
+    store_uncompressed_kernel=False,
+    method=None,
+    starting_l=1e-2,
+):
     r"""Perform regularized non-negative least-squares "fit" on self.
 
     Capable of solving for solution in 1 or 2 dimensions.
 
     We seek to minimize
     :math:`Q = \| Ax - b \|_2 + \|\lambda x\|_2`
-    in order to obtain solution vector :math:`x` subject to non-negativity constraint
-    given input matrix :math:`A`, the kernel, and input vector :math:`b`, the data.
+    in order to obtain solution vector :math:`x` subject to non-negativity
+    constraint given input matrix :math:`A`, the kernel, and input vector
+    :math:`b`, the data.
 
-    The first term assesses agreement between the fit :math:`Ax` and the data :math:`b`,
-    and the second term accounts for noise with the regularization parameter :math:`\lambda`
-    according to Tikhonov regularization.
+    The first term assesses agreement between the fit :math:`Ax` and the data
+    :math:`b`, and the second term accounts for noise with the regularization
+    parameter :math:`\lambda` according to Tikhonov regularization.
 
     To perform regularized minimization in 1 dimension, provide
     :str:`dimname_list`, :nddata:`newaxis_dict`, :function:`kernel_func`, and
-    regularization parameter `l`. One may set `l` to a :double: of the regularization
-    parameter of choice (found, for instance, through L-curve analysis) or
-    set `l` to :str:`BRD` to enable automatic selection of a regularization
-    parameter via the BRD algorithm - namely that described in Venkataramanan et al. 2002
-    but adapted for 1D case (DOI:10.1109/78.995059).
+    regularization parameter `l`. One may set `l` to a :double: of the
+    regularization parameter of choice (found, for instance, through L-curve
+    analysis) or set `l` to :str:`BRD` to enable automatic selection of a
+    regularization parameter via the BRD algorithm - namely that described in
+    Venkataramanan et al. 2002 but adapted for 1D case (DOI:10.1109/78.995059).
 
     To perform regularized minimization in 2 dimensions, set `l` to
     :str:`BRD` and provide a tuple of parameters :str:`dimname_list`,
@@ -133,8 +112,10 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
     parameter is supported in this 2 dimensional should an
     appropriate parameter be known.
 
-    See: `Wikipedia page on NNLS <https://en.wikipedia.org/wiki/Non-negative_least_squares>`_,
-    `Wikipedia page on Tikhonov regularization <https://en.wikipedia.org/wiki/Tikhonov_regularization>`_
+    See: `Wikipedia page on NNLS
+    <https://en.wikipedia.org/wiki/Non-negative_least_squares>`_,
+    `Wikipedia page on Tikhonov regularization
+    <https://en.wikipedia.org/wiki/Tikhonov_regularization>`_
 
     Parameters
     ==========
@@ -168,8 +149,8 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
         (in the example above, this would be something like
         ``lambda x,y: exp(-x*y)``)
 
-        For 2D, this must be a tuple or dictionary of functions -- the kernel is
-        the product of the two.
+        For 2D, this must be a tuple or dictionary of functions -- the kernel
+        is the product of the two.
     l : double (default 0) or str
         the regularization parameter :math:`lambda` -- if this is
         set to 0, the algorithm reverts to standard nnls.  If this
@@ -177,6 +158,9 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
         according to the BRD algorithm, either in 1-dimension or
         2-dimensions depending on presence of tuple synax
         (i.e., specifying more than 1 dimension).
+    method : callable, optional
+        Function to compute the regularized solution when ``l`` is not
+        ``"BRD"``.  Defaults to :func:`pyspecdata.nnls.nnls_regularized`.
 
     Returns
     =======
@@ -198,7 +182,7 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
     """
     # {{{ type checking
     demand_real(self.data)
-    if type(dimname_list) is str and type(newaxis_dict) is type(self):
+    if type(dimname_list) is str and isinstance(newaxis_dict, nddata_cls):
         # one dimensional with axis given as nddata
         dimname_list = [dimname_list]
         newaxis_dict = [newaxis_dict]
@@ -218,12 +202,16 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
     else:
         raise ValueError(
             strm(
-                "I didn't understand what you specified for the new axes (dimension:",
-                dimname_list,
-                "and new axes",
-                newaxis_dict,
+                "I didn't understand what you specified for the new axes"
+                " (dimension of type:",
+                type(dimname_list),
+                "and new axes type",
+                type(newaxis_dict),
+                "type of self is",
+                type(self),
             )
         )
+    logger.debug("argcheck done")
     # make sure axes are real
     for j in dimname_list:
         demand_real(
@@ -236,12 +224,14 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
         if j.getaxis(j.dimlabels[0]) is not None:
             demand_real(
                 j.getaxis(j.dimlabels[0]),
-                "(this message pertains to the new %s axis pulled from the second argument's axis)"
+                "(this message pertains to the new %s axis pulled from the"
+                " second argument's axis)"
                 % str(j.dimlabels[0]),
             )
         demand_real(
             j.data,
-            "(this message pertains to the new %s axis pulled from the second argument's data)"
+            "(this message pertains to the new %s axis pulled from the second"
+            " argument's data)"
             % str(j.dimlabels[0]),
         )
     if isinstance(kernel_func, tuple):
@@ -267,16 +257,14 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
     # equal to dimnames (length 1 for 1D and 2 for 2D)
     twoD = len(dimname_list) > 1
     fit_dimnames = [j.dimlabels[0] for j in newaxis_dict]
-    fit_axis_coords = [
-        j.getaxis(fit_dimnames[j_idx])
+    fit_axes = [
+        j
         for j_idx, j in enumerate(newaxis_dict)
         if j.getaxis(fit_dimnames[j_idx]) is not None
     ]
-    fit_axes = [
-        self.__class__(fit_axis_coords[j], fit_dimnames[j])
-        for j in range(len(dimname_list))
+    data_axes = [
+        self.fromaxis(dimname_list[j]) for j in range(len(dimname_list))
     ]
-    data_axes = [self.fromaxis(dimname_list[j]) for j in range(len(dimname_list))]
     # at this point, fit_axes and data_axes
     # are nddata objects that give the axis
     # coordinates in the fit and data
@@ -288,15 +276,13 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
     # inner dimensions, so we put out active dimensions on the inside
     # (last), so we want that order, but we want to save the final order
     # that we're going to want at the end
-    new_order = [fit_dimnames[dimname_list.index(j)]
-            if j in dimname_list else j
-            for j in self.dimlabels]
-    self.reorder(
-        dimname_list,
-        first=False
-    )
+    new_order = [
+        fit_dimnames[dimname_list.index(j)] if j in dimname_list else j
+        for j in self.dimlabels
+    ]
+    self.reorder(dimname_list, first=False)
     # }}}
-    logger.debug(strm("here is the order of the data",self.shape))
+    logger.debug(strm("here is the order of the data", self.shape))
     # }}}
     # {{{ construct the kernel
     # the kernel transforms from (columns) the "fit" dimension to (rows)
@@ -313,6 +299,8 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
                     kernels[j].dimlabels,
                     "and raw np.shape",
                     kernels[j].data.shape,
+                    "and data type",
+                    kernels[j].data.dtype,
                 )
                 for j in range(len(dimname_list))
             )
@@ -320,27 +308,29 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
     )
     # }}}
     if store_uncompressed_kernel:
-        self.set_prop('nnls_kernels_uncompressed',kernels)
+        self.set_prop("nnls_kernels_uncompressed", kernels)
     U, S, V = [[None] * len(dimname_list) for j in range(3)]
     s = [None] * len(dimname_list)
     for j in range(len(dimname_list)):
         U[j], S[j], V[j] = np.linalg.svd(kernels[j].data, full_matrices=False)
         logger.debug(
             strm(
-                f"the first few singular value are",
+                "the first few singular value are",
                 S[j][:4],
                 "the biggest is",
                 S[j][0],
                 "based on default_cut of",
                 default_cut,
                 "I'm going to cut out everything below",
-                S[j][0]*default_cut,
+                S[j][0] * default_cut,
             )
         )
-        s[j] = np.where(S[j] > default_cut * S[j][0])[0][-1] # JF changed this and following -- should be relative
+        s[j] = np.where(S[j] > default_cut * S[j][0])[0][
+            -1
+        ]  # JF changed this and following -- should be relative
         logger.debug(
             strm(
-                f"S{j+1} is",
+                f"S{j + 1} is",
                 S[j],
                 "so I'm going to",
                 s[j],
@@ -349,12 +339,15 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
             )
         )
     for j in range(len(dimname_list)):
-        U[j] = U[j][:, 0:s[j]]
-        S[j] = S[j][0:s[j]]
-        V[j] = V[j][0:s[j], :]
+        U[j] = U[j][:, 0 : s[j]]
+        S[j] = S[j][0 : s[j]]
+        V[j] = V[j][0 : s[j], :]
         S[j] = S[j] * np.eye(s[j])
         logger.debug(
-            strm("Compressed SVD of K[j]:", [x.shape for x in (U[j], S[j], V[j])])
+            strm(
+                "Compressed SVD of K[j]:",
+                [x.shape for x in (U[j], S[j], V[j])],
+            )
         )
     # {{{ compressing -- K are the compressed kernels (ΣV)
     K = [None] * len(dimname_list)
@@ -377,15 +370,13 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
         # sum across last and second from last and second to last
         # dimensions:
         # dot(a, b)[i,j,k,m] = sum(a[i,j,:] * b[k,:,m])
-        data_fornnls = U[0].T.dot(
-                self.data.dot(U[1])
-                )
+        data_fornnls = U[0].T.dot(self.data.dot(U[1]))
         # we want to smoosh the last two dimensions to get lex ordering
         # (here, if data has outer (earlier) dimensions, they are not
         # affected)
         data_fornnls = data_fornnls.reshape(
-                data_fornnls.shape[:-2] + (np.prod(data_fornnls.shape[-2:]),)
-                )
+            data_fornnls.shape[:-2] + (np.prod(data_fornnls.shape[-2:]),)
+        )
         # }}}
     else:
         K_alldims = S[0] @ V[0]
@@ -393,7 +384,16 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
         # [indirect × data] [data × SV] = [indirect × SV]
         # which is what we want
         data_fornnls = self.data.dot(U[0])
-        logger.debug(strm("data has shape",self.data.shape,"compressed to",data_fornnls.shape,"by U^T of",U[0].T.shape))
+        logger.debug(
+            strm(
+                "data has shape",
+                self.data.shape,
+                "compressed to",
+                data_fornnls.shape,
+                "by U^T of",
+                U[0].T.shape,
+            )
+        )
     # we are now ready to perform the regularization
     # along the innermost dimension, which is lex ordered where relevant!
     logger.debug(
@@ -405,15 +405,39 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
         )
     )
     if type(l) is str and l == "BRD":
-        if twoD:
-            factor = np.sqrt(np.prod(s))
-        else:
-            factor = None
-        retval, residual = this_nnls.nnls_regularized(
-            K_alldims, data_fornnls, l=mod_BRD(1.0, K_alldims, factor, data_fornnls)
+        # Automatic BRD parameter selection without stacking identity
+        # 1) Compute optimal f⃗ᵣ and λ via venk_BRD
+        logger.debug(
+            "about to run venk" + strm(K_alldims.shape, data_fornnls.shape)
         )
+        retval, l = venk_BRD(starting_l**2, K_alldims, data_fornnls)
+        self.set_prop("opt_alpha", l)
+        # 2) Compute residual: A·x − b (eq. 40)
+        residual = K_alldims.dot(retval) - data_fornnls
+        # Replace l with the chosen λ for downstream recording
     else:
-        retval, residual = this_nnls.nnls_regularized(K_alldims, data_fornnls, l=l)
+        if method is None:
+            method = this_nnls.nnls_regularized
+        logger.debug(
+            strm(
+                "I'm preparing to call",
+                getattr(method, "__name__", str(method)),
+                "with kernel dtype",
+                K_alldims.dtype,
+                "and shape",
+                K_alldims.shape,
+                "data dtype",
+                data_fornnls.dtype,
+                "and shape",
+                data_fornnls.shape,
+            )
+        )
+        result = method(K_alldims, data_fornnls, l=l)
+        if isinstance(result, tuple):
+            retval, residual = result
+        else:
+            retval = result
+            residual = K_alldims.dot(retval) - data_fornnls
     logger.debug(
         strm("coming back from fortran, residual type is", type(residual))
         + strm(residual.dtype if isinstance(residual, np.ndarray) else "")
@@ -422,7 +446,9 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
     if not np.isscalar(l):
         newshape.append(len(l))
     logger.debug(strm("test***", list(self.data.shape)[:-1]))
-    newshape += [self.shape[j] for j in self.dimlabels if j not in dimname_list] # any outer dimensions
+    newshape += [
+        self.shape[j] for j in self.dimlabels if j not in dimname_list
+    ]  # any outer dimensions
     newshape += [fit_axes[j].data.size for j in range(len(fit_dimnames))]
     logger.debug(
         strm(
@@ -450,7 +476,9 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
         axis_coords_dict[fit_dimnames[j]] = fit_axes[j].getaxis(
             fit_dimnames[j]
         )
-        axis_units_dict[fit_dimnames[j]] = None
+        axis_units_dict[fit_dimnames[j]] = fit_axes[j].get_units(
+            fit_dimnames[j]
+        )
         axis_coords_error_dict[fit_dimnames[j]] = None
     if not np.isscalar(l):
         self.dimlabels = ["lambda"] + self.dimlabels
@@ -458,21 +486,11 @@ def nnls(self, dimname_list, newaxis_dict, kernel_func, l=0, default_cut=1e-3,
         axis_units_dict["lambda"] = None
         axis_coords_error_dict["lambda"] = None
     self.data = retval
-    if not np.isscalar(residual):
-        # make the residual nddata as well
-        residual_nddata = self.shape.pop(fit_dimnames[0])
-        if twoD:
-            residual_nddata.pop(fit_dimnames[1])
-        residual_nddata = residual_nddata.alloc(dtype=residual.dtype)
-        residual_nddata.data[:] = residual[:]
-    else:
-        residual_nddata = residual
     # store the kernel and the residual as properties
     self.set_prop("nnls_kernel", K_alldims)
     for j in range(len(dimname_list)):
-        self.set_prop(f"s{j+1}", s[j])
-        self.set_prop(f"K{j+1}", K[j])
-    self.set_prop("nnls_residual", residual_nddata)
+        self.set_prop(f"s{j + 1}", s[j])
+        self.set_prop(f"K{j + 1}", K[j])
     self.axis_coords = self.fld(axis_coords_dict)
     self.axis_coords_units = self.fld(axis_units_dict)
     self.axis_coords_error = self.fld(axis_coords_error_dict)
